"""Models for categories."""

import json

import time
from iso8601 import parse_date
from iso8601.iso8601 import ParseError

from django.apps import apps
from django.conf import settings
from django.db import models

from simple_history.models import HistoricalRecords

from geokey.core.exceptions import InputError

from .managers import CategoryManager, FieldManager, LookupValueManager
from .base import STATUS, DEFAULT_STATUS


class Category(models.Model):
    """
    Defines the data structure of a certain type of features.
    """
    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)
    project = models.ForeignKey('projects.Project', related_name='categories')
    created_at = models.DateTimeField(auto_now_add=True)
    creator = models.ForeignKey(settings.AUTH_USER_MODEL)
    order = models.IntegerField(default=0)
    status = models.CharField(
        choices=STATUS,
        default=STATUS.active,
        max_length=20
    )
    display_field = models.ForeignKey(
        'categories.Field',
        null=True,
        related_name='display_field_of'
    )
    expiry_field = models.ForeignKey(
        'categories.Field',
        null=True,
        related_name='expiry_field_of'
    )
    default_status = models.CharField(
        choices=DEFAULT_STATUS,
        default=DEFAULT_STATUS.pending,
        max_length=20
    )
    colour = models.TextField(default='#0033ff')
    symbol = models.ImageField(upload_to='symbols', null=True, max_length=500)

    objects = CategoryManager()
    history = HistoricalRecords()

    class Meta:
        ordering = ['order']

    def reorder_fields(self, order):
        """
        Changes the order in which fields are displayed on client side.

        Parameters
        -------
        order : List
            IDs of fields, ordered according to new display order
        """
        fields_to_save = []
        for idx, field_id in enumerate(order):
            field = self.fields.get(pk=field_id)
            field.order = idx
            fields_to_save.append(field)

        for field in fields_to_save:
            field.save()

    def get_query(self, rule):
        """
        Returns the SQL where clause for the category. It combines the where
        clause parts of each field in the category.

        Returns
        -------
        str
            SQL where clause for the rule
        """
        queries = ['(category_id = %s)' % self.id]

        if 'min_date' in rule:
            queries.append('("contributions_observation".created_at >= to_date(\'' +
                           rule['min_date'] +
                           '\', \'YYYY-MM-DD HH24:MI\'))')

        if 'max_date' in rule:
            queries.append('("contributions_observation".created_at <= to_date(\'' +
                           rule['max_date'] +
                           '\', \'YYYY-MM-DD HH24:MI\'))')

        for key in rule:
            if key not in ['min_date', 'max_date']:
                try:
                    field = self.fields.get_subclass(key=key)
                    queries.append(field.get_filter(rule[key]))
                except Field.DoesNotExist:
                    pass

        return '(%s)' % ' AND '.join(queries)

    def delete(self):
        """
        Deletes the category by setting its status to deleted.

        Notes
        -----
        It also deletes all contributions of that category.
        """
        from geokey.contributions.models import Observation
        Observation.objects.filter(category=self).delete()

        groups = self.project.usergroups.all()
        for usergroup in groups:
            if usergroup.filters is not None:
                f = usergroup.filters.pop(str(self.id), None)
                if f is not None:
                    usergroup.save()

        self.status = STATUS.deleted
        self.save()


class Field(models.Model):
    """
    A Field defines data type of one characterictic of an obesrvation. Used to
    create forms of user interfaces and to validate user inputs.
    """
    name = models.CharField(max_length=100)
    key = models.CharField(max_length=103)
    description = models.TextField(null=True, blank=True)
    required = models.BooleanField(default=False)
    category = models.ForeignKey('Category', related_name='fields')
    order = models.IntegerField(default=0)
    status = models.CharField(
        choices=STATUS,
        default=STATUS.active,
        max_length=20
    )

    objects = FieldManager()

    class Meta:
        unique_together = ('key', 'category')

    @classmethod
    def create(cls, name, key, description, required, category, field_type):
        """
        Creates a new field based on the field type provided.

        Parameters
        ----------
        name : str
            Name of the field. Used for displaying labels.
        key : str
            Key of the field. Used in API to assign values to a field.
        description : str
            Long description providing further details about the field.
        required : Boolean
            Indicates if the field is required
        category : geokey.categories.models.Category
            The category this field is assigned to.
        field_type : str
            Type of the field. Must be one of:
                - TextField
                - NumericField
                - DateField
                - DateTimeField
                - TimeField
                - LookupField
                - MultipleLookupfield

        Returns
        -------
        geokey.categories.models.Field
            Intance of the newly created field
        """
        model_class = apps.get_model('categories', field_type)
        order = category.fields.count()

        field = model_class.objects.create(
            name=name,
            key=key,
            description=description,
            required=required,
            category=category,
            order=order
        )
        field.save()

        if order == 0:
            category.display_field = field
            category.save()

        return field

    @classmethod
    def get_field_types(cls):
        """
        Returns the names of the subclasses of Field.

        Returns
        -------
        List
            The names of available field types.
        """
        return cls.__subclasses__()

    def validate_input(self, value):
        """
        Validates the given value against the field definition.
        @abstractmethod
        """
        raise NotImplementedError(
            'The method `validate_input` has not been implemented for this '
            'child class of Field.'
        )

    def validate_required(self, value):
        """
        Validates input value against required status.
        Raises an `InputError` if no value has been provided.
        """
        if self.status == STATUS.active and self.required and (value is None):
            raise InputError('The field %s is required.' % self.name)

    @property
    def fieldtype(self):
        """
        Returns the class name of the field instance

        Returns
        -------
        str
            The class name of the field instance
        """
        return self.__class__.__name__

    @property
    def type_name(self):
        """
        Returns the type name of the field instance. This is a human-readable
        name that can be used in user interfaces, e.g. Date and Time for
        DateTimeField.
        @abstractmethod
        """
        raise NotImplementedError(
            'The property `type_name` has not been implemented for this '
            'subclass of Field.'
        )

    def get_filter(self, rule):
        """
        Returns an SQL where clause that can be used to filter contributions in
        data groupings
        @abstractmethod

        Parameter
        ---------
        rule : str or list or dict
            Depending on the field type, this provides the values the filter
            should be built against

        Return
        ------
        str
            The where-clause that can be used in SQL queries.
        """
        raise NotImplementedError(
            'The method `filter` has not been implemented for this '
            'subclass of Field.'
        )

    def delete(self):
        """
        Deletes the field. It also removes the field from the filters attached
        to substes and user groups.
        """
        for subset in self.category.project.subsets.all():
            subset.remove_filter_field(self)

        for usergroup in self.category.project.usergroups.all():
            usergroup.remove_filter_field(self)

        super(Field, self).delete()


class TextField(Field):
    """
    A field for character strings.
    """
    textarea = models.BooleanField(default=False)
    maxlength = models.IntegerField(blank=True, null=True)

    def validate_required(self, value):
        """
        Validate the given value agaist required status. Checks if value is
        not None and has at least one character.

        Parameters
        ----------
        value : str
            The value to be validated

        Notes
        -----
        Raises InputError if no value is provided
        """
        if isinstance(value, str) or isinstance(value, unicode):
            value = value.encode('utf-8')

        if self.status == STATUS.active and self.required and (
                value is None or len(str(value)) == 0):
            raise InputError('The field %s is required.' % self.name)

    def validate_input(self, value):
        """
        Validates if the given value is a valid input for the TextField.
        Checks if the value is required and if maxlength constraint is met.

        Parameters
        ----------
        value : str
            The value to be validated

        Notes
        -----
        Raises InputError if an invalid value is provided
        """
        self.validate_required(value)

        if value is not None:
            if self.maxlength is not None and len(value) > self.maxlength:
                raise InputError('The input provided for text field %s '
                                 'contains too many characters.' % self.name)

    @property
    def type_name(self):
        """
        Returns a human readable name of the field.

        Return
        ------
        str
            The name of the field
        """
        return 'Text'

    def get_filter(self, rule):
        """
        Returns the SQL where clause for the given field based on the rule.
        Used to filter data for user groups and subsets.

        Parameter
        ---------
        rule : str
            A keyword that needs to matched for the filter to apply.

        Return
        ------
        str
            SQL where-clause
        """
        return ('((properties ->> \'' + self.key + '\') '
                'ILIKE \'%%' + rule + '%%\')')


class NumericField(Field):
    """
    A field for numeric values.
    """
    minval = models.FloatField(blank=True, null=True)
    maxval = models.FloatField(blank=True, null=True)

    def validate_input(self, value):
        """
        Validates if the given value is a valid input for the NumerField.
        Checks if a value of type number has been provided or if a value of
        type String has been provided that can be successfully converted to a
        Float value. Then checks if the value is between bounds of minval and
        maxval.

        Parameters
        ----------
        value : str
            The value to be validated

        Notes
        -----
        Raises InputError if an invalid value is provided
        """
        if isinstance(value, (str, unicode)) and len(value) == 0:
            value = None

        self.validate_required(value)

        if value is not None:
            if isinstance(value, (str, unicode)):
                try:
                    value = float(value) if '.' in value else int(value)
                except ValueError:
                    raise InputError(
                        'The value provided for field %s is not a number.' %
                        self.name
                    )

            if isinstance(value, (int, long, float, complex)):
                if self.minval and self.maxval and (
                        not (value >= self.minval) and (value <= self.maxval)):
                    raise InputError('The value provided for field %s must be '
                                     ' greater than %s and lower than %s.'
                                     % (self.name, self.minval, self.maxval))

                else:
                    if self.minval and (not (value >= self.minval)):
                        raise InputError('The value provided for field %s must'
                                         ' be greater than %s.'
                                         % (self.name, self.minval))

                    if self.maxval and (not (value <= self.maxval)):
                        raise InputError('The value provided for field %s must'
                                         ' be lower than %s.'
                                         % (self.name, self.maxval))

    @property
    def type_name(self):
        """
        Returns a human readable name of the field.

        Return
        ------
        str
            The name of the field
        """
        return 'Numeric'

    def get_filter(self, rule):
        """
        Returns the SQL where clause for the given field based on the rule.
        Used to filter data for user groups and subsets.

        Parameter
        ---------
        rule : dict
            Contains either minimum and maximum value for the filer:
            {minval: 1, maxval: 10}

        Return
        ------
        str
            SQL where-clause
        """
        minval = rule.get('minval')
        maxval = rule.get('maxval')

        if minval is not None and maxval is not None:
            return ('(cast(properties ->> \'%s\' as double precision) >= %s) '
                    'AND (cast(properties ->> \'%s\' as double precision) <= '
                    '%s)' % (self.key, minval, self.key, maxval))
        else:
            if minval is not None:
                return ('(cast(properties ->> \'%s\' as double '
                        'precision) >= %s)' % (self.key, minval))

            if maxval is not None:
                return ('(cast(properties ->> \'%s\' as double '
                        'precision) <= %s)' % (self.key, maxval))


class DateTimeField(Field):
    """
    A field for storing dates and times.
    """

    def validate_required(self, value):
        """
        Validate the given value agaist required status. Checks if value is
        not None and has at least one character.

        Parameters
        ----------
        value : str or None
            The value to be validated

        Notes
        -----
        Raises InputError if no value is provided
        """
        if (self.status == STATUS.active and
                self.required and (value is None or len(value) == 0)):
            raise InputError('The field %s is required.' % self.name)

    def validate_input(self, value):
        """
        Checks if the provided value is a valid and ISO8601 compliant date
        string.

        Parameters
        ----------
        value : str
            The value to be validated

        Notes
        -----
        Raises InputError if an invalid value is provided
        """
        self.validate_required(value)
        if value is not None:
            try:
                parse_date(value)
            except ParseError:
                raise InputError('The value for DateField %s is not a '
                                 'valid date. Please provide date and time as '
                                 'YYYY-MM-DD HH:MM' % self.name)

    @property
    def type_name(self):
        """
        Returns a human readable name of the field.

        Return
        ------
        str
            The name of the field
        """
        return 'Date and Time'

    def get_filter(self, rule):
        """
        Returns the SQL where clause for the given field based on the rule.
        Used to filter data for user groups and subsets.

        Parameter
        ---------
        rule : dict
            Contains either minimum and maximum value for the filer:
            {minval: '2015-10-01', maxval: '2015-10-31'}

        Return
        ------
        str
            SQL where-clause
        """
        minval = rule.get('minval')
        maxval = rule.get('maxval')

        if minval is not None and maxval is not None:
            return ('(to_date(properties ->> \'%s\', \'YYYY-MM-DD HH24:MI\') '
                    '>= to_date(\'%s\', \'YYYY-MM-DD HH24:MI\')) AND '
                    '(to_date(properties ->> \'%s\', \'YYYY-MM-DD HH24:MI\') '
                    '<= to_date(\'%s\', \'YYYY-MM-DD HH24:MI\'))' %
                    (self.key, minval, self.key, maxval))
        else:
            if minval is not None:
                return ('(to_date(properties ->> \'%s\', \'YYYY-MM-DD '
                        'HH24:MI\') >= to_date(\'%s\', \'YYYY-MM-DD HH24:MI\''
                        '))' % (self.key, minval))

            if maxval is not None:
                return ('(to_date(properties ->> \'%s\', \'YYYY-MM-DD '
                        'HH24:MI\') <= to_date(\'%s\', \'YYYY-MM-DD HH24:MI\''
                        '))' % (self.key, maxval))


class DateField(Field):
    """
    A field for storing dates.
    """
    def validate_required(self, value):
        """
        Validate the given value agaist required status. Checks if value is
        not None and has at least one character.

        Parameters
        ----------
        value : str or None
            The value to be validated

        Notes
        -----
        Raises InputError if no value is provided
        """
        if (self.status == STATUS.active and self.required and
                (value is None or len(value) == 0)):
            raise InputError('The field %s is required.' % self.name)

    def validate_input(self, value):
        """
        Checks if the provided value is a valid and ISO8601 compliant date
        string.

        Parameters
        ----------
        value : str
            The value to be validated

        Notes
        -----
        Raises InputError if an invalid value is provided
        """
        self.validate_required(value)
        if value is not None:
            try:
                parse_date(value)
            except ParseError:
                raise InputError('The value for DateField %s is not a '
                                 'valid date. Please provide date as '
                                 'YYYY-MM-DD.' % self.name)

    @property
    def type_name(self):
        """
        Returns a human readable name of the field.

        Return
        ------
        str
            The name of the field
        """
        return 'Date'

    def get_filter(self, rule):
        """
        Returns the SQL where clause for the given field based on the rule.
        Used to filter data for user groups and subsets.

        Parameter
        ---------
        rule : dict
            Contains either minimum and maximum value for the filer:
            {minval: '2015-10-01 10:00', maxval: '2015-10-31 15:00'}

        Return
        ------
        str
            SQL where-clause
        """
        minval = rule.get('minval')
        maxval = rule.get('maxval')

        if minval is not None and maxval is not None:
            return ('(to_date(properties ->> \'%s\', \'YYYY-MM-DD\') '
                    '>= to_date(\'%s\', \'YYYY-MM-DD\')) AND '
                    '(to_date(properties ->> \'%s\', \'YYYY-MM-DD\') '
                    '<= to_date(\'%s\', \'YYYY-MM-DD\'))' %
                    (self.key, minval, self.key, maxval))
        else:
            if minval is not None:
                return ('(to_date(properties ->> \'%s\', \'YYYY-MM-DD'
                        '\') >= to_date(\'%s\', \'YYYY-MM-DD\''
                        '))' % (self.key, minval))

            if maxval is not None:
                return ('(to_date(properties ->> \'%s\', \'YYYY-MM-DD'
                        '\') <= to_date(\'%s\', \'YYYY-MM-DD\''
                        '))' % (self.key, maxval))


class TimeField(Field):
    @property
    def type_name(self):
        """
        Returns a human readable name of the field.

        Return
        ------
        str
            The name of the field
        """
        return 'Time'

    def validate_required(self, value):
        """
        Validate the given value agaist required status. Checks if value is
        not None and has at least one character.

        Parameters
        ----------
        value : str or None
            The value to be validated

        Notes
        -----
        Raises InputError if no value is provided
        """
        if (self.status == STATUS.active and self.required and
                (value is None or len(value) == 0)):
            raise InputError('The field %s is required.' % self.name)

    def validate_input(self, value):
        """
        Checks if the provided value is a matches the patterns HH:mm

        Parameters
        ----------
        value : str
            The value to be validated

        Notes
        -----
        Raises InputError if an invalid value is provided
        """
        self.validate_required(value)
        if value is not None:
            try:
                time.strptime(value, '%H:%M')
            except ValueError:
                raise InputError('The value for TimeField %s is not a '
                                 'valid time. Please provide time as '
                                 'HH:MM' % self.name)

    def get_filter(self, rule):
        """
        Returns the SQL where clause for the given field based on the rule.
        Used to filter data for a user groups and subsets.

        Parameter
        ---------
        rule : dict
            Contains either minimum and maximum value for the filer:
            {minval: '10:00', maxval: '15:00'}

        Return
        ------
        str
            SQL where-clause
        """
        minval = rule.get('minval')
        maxval = rule.get('maxval')

        if minval is not None and maxval is not None:
            if time.strptime(minval, '%H:%M') > time.strptime(maxval, '%H:%M'):
                return ('((properties ->> \'%s\')::time >= \'%s\'::time) OR '
                        '((properties ->> \'%s\')::time <= \'%s\'::time)' %
                        (self.key, minval, self.key, maxval))
            else:
                return ('((properties ->> \'%s\')::time >= \'%s\'::time) AND '
                        '((properties ->> \'%s\')::time <= \'%s\'::time)' %
                        (self.key, minval, self.key, maxval))
        else:
            if minval is not None:
                return ('((properties ->> \'%s\')::time >= \'%s\'::time)' %
                        (self.key, minval))

            if maxval is not None:
                return ('((properties ->> \'%s\')::time <= \'%s\'::time)' %
                        (self.key, maxval))


class LookupField(Field):
    """
    A lookup value is a special kind of field the provides an pre-defined
    number of values as valid input values.
    """
    def validate_input(self, value):
        """
        Checks if the provided value matches the ID of one of the field's
        lookupvalues.

        Parameters
        ----------
        value : int
            The value to be validated

        Notes
        -----
        Raises InputError if an invalid value is provided
        """
        self.validate_required(value)

        valid = False

        if value is not None:
            try:
                value = int(value)
                for lookupvalue in self.lookupvalues.all():
                    if lookupvalue.id == value:
                        valid = True
            except ValueError:
                pass
        else:
            valid = True

        if not valid:
            raise InputError('The value for lookup field %s is not an '
                             'accepted value for the field.' % self.name)

    @property
    def type_name(self):
        """
        Returns a human readable name of the field.

        Return
        ------
        str
            The name of the field
        """
        return 'Select box'

    def get_filter(self, rule):
        """
        Returns the SQL where clause for the given field based on the rule.
        Used to filter data for a user groups and subsets.

        Parameter
        ---------
        rule : List
            IDs of LookupValues that need to matched in order for the filter
            to apply.

        Return
        ------
        str
            SQL where-clause
        """
        return ('((properties ->> \'%s\')::int IN (%s))' %
                (self.key, ','.join(str(x) for x in rule)))


class LookupValue(models.Model):
    """
    Stores a single lookup value.
    """
    name = models.CharField(max_length=100)
    symbol = models.ImageField(upload_to='symbols', null=True, max_length=500)
    field = models.ForeignKey(LookupField, related_name='lookupvalues')
    status = models.CharField(
        choices=STATUS,
        default=STATUS.active,
        max_length=20
    )

    objects = LookupValueManager()

    class Meta:
        ordering = ['id']

    def delete(self):
        """
        Deletes the value by settings its status to `deleted`
        """
        self.status = STATUS.inactive
        self.save()


class MultipleLookupField(Field):
    def validate_input(self, provided_vals):
        """
        Checks if the provided value matches the ID of one of the field's
        lookupvalues.

        Parameters
        ----------
        value : int
            The value to be validated

        Notes
        -----
        Raises InputError if an invalid value is provided
        """
        self.validate_required(provided_vals)

        valid = True

        if provided_vals is not None:
            if isinstance(provided_vals, (str, unicode)):
                provided_vals = json.loads(provided_vals)

            accepted_values = [value.id for value in self.lookupvalues.all()]
            intersection = [
                val for val in provided_vals if val in accepted_values
            ]

            valid = len(intersection) == len(provided_vals)

        if not valid:
            raise InputError('One or more values for the multiple select '
                             'field %s is not an accepted value for the '
                             'field.' % self.name)

    @property
    def type_name(self):
        """
        Returns a human readable name of the field.

        Return
        ------
        str
            The name of the field
        """
        return 'Multiple select'

    def get_filter(self, rule):
        """
        Returns the SQL where clause for the given field based on the rule.
        Used to filter data for user groups and subsets.

        Parameter
        ---------
        rule : List
            IDs of LookupValues that need to matched in order for the filter
            to apply.

        Return
        ------
        str
            SQL where-clause
        """
        return ('(regexp_split_to_array(btrim(properties ->> \'%s\', \'[]\'),'
                ' \',\')::int[] && ARRAY[%s])' % (self.key, ', '.join(rule)))


class MultipleLookupValue(models.Model):
    """
    Stores a multiple lookup value.
    """
    name = models.CharField(max_length=100)
    symbol = models.ImageField(upload_to='symbols', null=True, max_length=500)
    field = models.ForeignKey(MultipleLookupField, related_name='lookupvalues')
    status = models.CharField(
        choices=STATUS,
        default=STATUS.active,
        max_length=20
    )

    objects = LookupValueManager()

    class Meta:
        ordering = ['id']

    def delete(self):
        """
        Deletes the value by settings its status to `deleted`
        """
        self.status = STATUS.inactive
        self.save()



"""Public API URLs."""

from django.conf.urls import url

from geokey.core.views import InfoAPIView

from geokey.projects import views as project_views
from geokey.categories import views as category_views

from geokey.contributions.views import observations, comments, locations, media
from geokey.users.views import UserAPIView, ChangePasswordView


urlpatterns = [
    # ###########################
    # CORE
    # ###########################
    url(
        r'^info/$',
        InfoAPIView.as_view(),
        name='info'),

    # ###########################
    # USERS
    # ###########################
    url(
        r'^user/$',
        UserAPIView.as_view(),
        name='user_api'),
    url(
        r'^user/password/change/$',
        ChangePasswordView.as_view(),
        name='changepassword'),

    # ###########################
    # PROJECTS
    # ###########################
    url(
        r'^projects/$',
        project_views.Projects.as_view(),
        name='project'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/$',
        project_views.SingleProject.as_view(),
        name='project_single'),

    # ###########################
    # CATEGORIES
    # ###########################
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'categories/(?P<category_id>[0-9]+)/$',
        category_views.SingleCategory.as_view(),
        name='category'),

    # ###########################
    # CONTRIBUTIONS
    # ###########################
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'contributions/$',
        observations.ProjectObservations.as_view(),
        name='project_observations'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'contributions/(?P<observation_id>[0-9]+)/$',
        observations.SingleAllContributionAPIView.as_view(),
        name='project_single_observation'),

    # ###########################
    # LOCATIONS
    # ###########################
    url(
        r'^projects/(?P<project_id>[0-9]+)'
        r'/locations/$',
        locations.LocationsAPIView.as_view(),
        name='project_locations'),
    url(
        r'^projects/(?P<project_id>[0-9]+)'
        r'/locations/(?P<location_id>[0-9]+)/$',
        locations.SingleLocationAPIView.as_view(),
        name='project_single_location'),

    # ###########################
    # COMMENTS
    # ###########################
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'contributions/(?P<contribution_id>[0-9]+)/'
        r'comments/$',
        comments.CommentsAPIView.as_view(),
        name='project_comments'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'contributions/(?P<contribution_id>[0-9]+)/'
        r'comments/(?P<comment_id>[0-9]+)/$',
        comments.SingleCommentAPIView.as_view(),
        name='project_single_comment'),

    # ###########################
    # MEDIA
    # ###########################
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'contributions/(?P<contribution_id>[0-9]+)/'
        r'media/$',
        media.MediaAPIView.as_view(),
        name='project_media'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'contributions/(?P<contribution_id>[0-9]+)/'
        r'media/(?P<file_id>[0-9]+)/$',
        media.SingleMediaAPIView.as_view(),
        name='project_single_media'),
]



"""Tests for URLs."""

from django.test import TestCase
from django.core.urlresolvers import reverse, resolve

from geokey.core.views import InfoAPIView


class UrlsTest(TestCase):
    """Test all URLs."""

    def test_reverse_info(self):
        """Test reverser for info API."""
        self.assertEqual(reverse('api:info'), '/api/info/')

    def test_resolve_subset_list(self):
        """Test resolver for info API."""
        resolved = resolve('/api/info/')
        self.assertEqual(resolved.func.__name__, InfoAPIView.__name__)



"""Views for comments of categories."""

from django.core.exceptions import PermissionDenied

from rest_framework import status
from rest_framework.views import APIView
from rest_framework.response import Response

from geokey.core.decorators import handle_exceptions_for_ajax
from geokey.core.exceptions import MalformedRequestData
from geokey.users.models import User

from .base import SingleAllContribution
from ..models import Comment
from ..serializers import CommentSerializer


class CommentAbstractAPIView(APIView):
    """Abstract class for comments."""

    def get_user(self, request):
        """
        Get user of a request.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.

        Returns
        -------
        geokey.users.models.User
            User of a request.
        """
        user = request.user

        if user.is_anonymous():
            user = User.objects.get(display_name='AnonymousUser')

        return user

    def get_comment(self, contribution, comment_id):
        """
        Get comment of a contribution.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        contribution : geokey.contributions.models.Observation
            Contribution to retrieve the comment from.
        comment_id : int
            Identifies the comment in the database.

        Returns
        -------
        geokey.contributions.models.Comment
            Comment of a contribution.
        """
        return contribution.comments\
            .select_related('creator')\
            .prefetch_related('responses')\
            .get(pk=comment_id)

    def get_list_and_respond(self, request, contribution):
        """
        Respond to a GET request with a list of all comments.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        contribution : geokey.contributions.models.Observation
            Contribution the comments are requested for.

        Returns
        -------
        rest_framework.response.Respones
            Contains the serialized comments.
        """
        serializer = CommentSerializer(
            contribution.comments.filter(respondsto=None),
            many=True,
            context={'user': self.get_user(request)}
        )
        return Response(serializer.data, status=status.HTTP_200_OK)

    def create_and_respond(self, request, contribution):
        """
        Respond to a POST request by creating a comment.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        contribution : geokey.contributions.models.Observation
            Contribution the comment should be added to.

        Returns
        -------
        rest_framework.response.Respone
            Contains the serialized comment.

        Raises
        ------
        MalformedRequestData
            When contribution is a draft or comment does not belong to the
            contribution.
        """
        user = self.get_user(request)

        if contribution.status == 'draft':
            raise MalformedRequestData(
                'This contribution is a draft. You may not comment on drafts.'
            )

        respondsto = request.data.get('respondsto') or None
        if respondsto:
            try:
                respondsto = contribution.comments.get(pk=respondsto)
            except Comment.DoesNotExist:
                raise MalformedRequestData(
                    'The comment you try to respond to is not a comment '
                    'to the contribution.'
                )

        review_status = request.data.get('review_status') or None
        if review_status == 'open' and contribution.status != 'review':
            contribution.update(None, user, status='review')

        comment = Comment.objects.create(
            text=request.data.get('text'),
            commentto=contribution,
            respondsto=respondsto,
            creator=user,
            review_status=review_status
        )

        serializer = CommentSerializer(comment, context={'user': user})
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def update_and_respond(self, request, contribution, comment):
        """
        Respond to a PATCH request by updating the comment.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        contribution : geokey.contributions.models.Observation
            Contribution of a comment.
        comment : geokey.contributions.models.Comment
            Comment to be updated.

        Returns
        -------
        rest_framework.response.Respone
            Contains the serialized comment.

        Raises
        ------
        PermissionDenied
            When user is not allowed to update the comment.
        """
        user = self.get_user(request)
        data = request.data

        is_owner = comment.creator == user
        can_moderate = contribution.project.can_moderate(user)

        if is_owner or can_moderate:
            if data.get('review_status') == 'resolved' and not can_moderate:
                raise PermissionDenied(
                    'You are not a project moderator and therefore not '
                    'eligable to resolve this comment.'
                )

            serializer = CommentSerializer(
                comment,
                data=data,
                partial=True,
                context={'user': user}
            )

            if serializer.is_valid():
                serializer.save()

                if (not contribution.comments.filter(
                        review_status='open').exists()):
                    contribution.update(None, user, status='active')

                return Response(serializer.data)
            else:
                return Response(
                    serializer.errors,
                    status=status.HTTP_400_BAD_REQUEST
                )
        else:
            raise PermissionDenied(
                'You are neither the author if this comment nor a project '
                'moderator and therefore not eligable to edit it.'
            )

    def delete_and_respond(self, request, contribution, comment):
        """
        Respond to a DELETE request by deleting the comment.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        contribution : geokey.contributions.models.Observation
            Contribution of a comment.
        comment : geokey.contributions.models.Comment
            Comment to be deleted.

        Returns
        -------
        rest_framework.response.Respone
            Empty response indicating success.

        Raises
        ------
        PermissionDenied
            When user is not allowed to delete the comment.
        """
        user = self.get_user(request)

        if comment.creator == user or contribution.project.can_moderate(user):
            comment.delete()

            if (contribution.status == 'review' and
                    not contribution.comments.filter(
                        review_status='open').exists()):
                contribution.update(None, user, status='active')

            return Response(status=status.HTTP_204_NO_CONTENT)
        else:
            raise PermissionDenied(
                'You are neither the author if this comment nor a project '
                'moderator and therefore not eligable to delete it.'
            )


class CommentsAPIView(SingleAllContribution, CommentAbstractAPIView):
    """Public API for all comments."""

    @handle_exceptions_for_ajax
    def get(self, request, project_id, contribution_id):
        """
        Handle GET request.

        Return a list of all comments of the contribution.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        project_id : int
            Identifies the project in the database.
        contribution_id : int
            Identifies the contribution in the database.

        Returns
        -------
        rest_framework.response.Respone
            Contains the serialised comments.
        """
        contribution = self.get_contribution(
            request.user,
            project_id,
            contribution_id
        )

        return self.get_list_and_respond(request, contribution)

    @handle_exceptions_for_ajax
    def post(self, request, project_id, contribution_id):
        """
        Handle POST request.

        Add a new comment to the contribution.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        project_id : int
            Identifies the project in the database.
        contribution_id : int
            Identifies the contribution in the database.

        Returns
        -------
        rest_framework.response.Respone
            Contains the serialised comment.
        """
        contribution = self.get_contribution(
            request.user,
            project_id,
            contribution_id
        )

        return self.create_and_respond(request, contribution)


class SingleCommentAPIView(SingleAllContribution, CommentAbstractAPIView):
    """Public API for a single comment."""

    @handle_exceptions_for_ajax
    def patch(self, request, project_id, contribution_id, comment_id):
        """
        Handle PATCH request.

        Update the comment.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        project_id : int
            Identifies the project in the database.
        contribution_id : int
            Identifies the contribution in the database.
        comment_id : int
            Identifies the comment in the database.

        Returns
        -------
        rest_framework.response.Respone
            Contains the serialized comment.
        """
        contribution = self.get_contribution(
            request.user,
            project_id,
            contribution_id
        )
        comment = self.get_comment(contribution, comment_id)

        return self.update_and_respond(request, contribution, comment)

    @handle_exceptions_for_ajax
    def delete(self, request, project_id, contribution_id, comment_id):
        """
        Handle DELETE request.

        Delete the comment.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        project_id : int
            Identifies the project in the database.
        contribution_id : int
            Identifies the contribution in the database.
        comment_id : int
            Identifies the comment in the database.

        Returns
        -------
        rest_framework.response.Respone
            Empty response indicating success.
        """
        contribution = self.get_contribution(
            request.user,
            project_id,
            contribution_id
        )
        comment = self.get_comment(contribution, comment_id)

        return self.delete_and_respond(request, contribution, comment)



"""Tests for logger: model User."""

from django.test import TestCase

from geokey.core.models import LoggerHistory
from geokey.users.tests.model_factories import UserFactory


class LogUserTest(TestCase):
    """Test model User."""

    def setUp(self):
        """Set up test."""
        self.user = UserFactory.create()

    def test_log_create(self):
        """Test when user gets created."""
        log_count_init = LoggerHistory.objects.count()
        UserFactory.create()
        self.assertEqual(LoggerHistory.objects.count(), log_count_init)

    def test_log_update_display_name(self):
        """Test when display name changes."""
        log_count_init = LoggerHistory.objects.count()
        self.user.display_name = '%s UPDATED' % self.user.display_name
        self.user.save()
        self.assertEqual(LoggerHistory.objects.count(), log_count_init)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


def update_count(apps, schema_editor):
    Observation = apps.get_model('contributions', 'Observation')

    for o in Observation.objects.all():
        o.num_media = o.files_attached.exclude(status='deleted').count()
        o.num_comments = o.comments.exclude(status='deleted').count()
        o.save()


class Migration(migrations.Migration):

    dependencies = [
        ('contributions', '0011_auto_20150527_1255'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicalobservation',
            name='num_comments',
            field=models.IntegerField(default=0),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='historicalobservation',
            name='num_media',
            field=models.IntegerField(default=0),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='observation',
            name='num_comments',
            field=models.IntegerField(default=0),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='observation',
            name='num_media',
            field=models.IntegerField(default=0),
            preserve_default=True,
        ),
        migrations.RunPython(update_count)
    ]



"""Model factories used for tests of categories."""

import factory

from geokey.core.tests.helpers.image_helpers import get_image
from geokey.users.tests.model_factories import UserFactory
from geokey.projects.tests.model_factories import ProjectFactory

from ..models import (
    Category, TextField, NumericField, DateTimeField, DateField, TimeField,
    LookupField, LookupValue, Field, MultipleLookupField, MultipleLookupValue
)


class CategoryFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Category

    creator = factory.SubFactory(UserFactory)
    name = factory.Sequence(lambda n: 'category %s' % n)
    description = factory.LazyAttribute(lambda o: '%s description' % o.name)
    project = factory.SubFactory(ProjectFactory)
    status = 'active'


class FieldFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Field

    name = factory.Sequence(lambda n: 'field %s' % n)
    key = factory.Sequence(lambda n: 'field_%s' % n)
    description = factory.LazyAttribute(lambda o: '%s description' % o.name)
    category = factory.SubFactory(CategoryFactory)
    status = 'active'
    required = False
    order = 0


class TextFieldFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = TextField

    name = factory.Sequence(lambda n: 'textfield %s' % n)
    key = factory.Sequence(lambda n: 'textfield_%s' % n)
    description = factory.LazyAttribute(lambda o: '%s description' % o.name)
    category = factory.SubFactory(CategoryFactory)
    status = 'active'
    required = False


class NumericFieldFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = NumericField

    name = factory.Sequence(lambda n: 'numericfield %s' % n)
    key = factory.Sequence(lambda n: 'numericfield_%s' % n)
    description = factory.LazyAttribute(lambda o: '%s description' % o.name)
    category = factory.SubFactory(CategoryFactory)
    status = 'active'
    required = False


class DateTimeFieldFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = DateTimeField

    name = factory.Sequence(lambda n: 'datetimefield %s' % n)
    key = factory.Sequence(lambda n: 'datetimefield_%s' % n)
    description = factory.LazyAttribute(lambda o: '%s description' % o.name)
    category = factory.SubFactory(CategoryFactory)
    status = 'active'
    required = False


class DateFieldFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = DateField

    name = factory.Sequence(lambda n: 'datefield %s' % n)
    key = factory.Sequence(lambda n: 'datefield_%s' % n)
    description = factory.LazyAttribute(lambda o: '%s description' % o.name)
    category = factory.SubFactory(CategoryFactory)
    status = 'active'
    required = False


class TimeFieldFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = TimeField

    name = factory.Sequence(lambda n: 'timefield %s' % n)
    key = factory.Sequence(lambda n: 'timefield_%s' % n)
    description = factory.LazyAttribute(lambda o: '%s description' % o.name)
    category = factory.SubFactory(CategoryFactory)
    status = 'active'
    required = False


class LookupFieldFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = LookupField

    name = factory.Sequence(lambda n: 'lookupfield %s' % n)
    key = factory.Sequence(lambda n: 'lookupfield_%s' % n)
    description = factory.LazyAttribute(lambda o: '%s description' % o.name)
    category = factory.SubFactory(CategoryFactory)
    status = 'active'
    required = False


class LookupValueFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = LookupValue

    name = factory.Sequence(lambda n: 'lookupvalue %s' % n)
    symbol = get_image(file_name='test_lookup_value_symbol.png')
    field = factory.SubFactory(LookupFieldFactory)
    status = 'active'


class MultipleLookupFieldFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MultipleLookupField

    name = factory.Sequence(lambda n: 'multiplelookupfield %s' % n)
    key = factory.Sequence(lambda n: 'multiplelookupfield_%s' % n)
    description = factory.LazyAttribute(lambda o: '%s description' % o.name)
    category = factory.SubFactory(CategoryFactory)
    status = 'active'
    required = False


class MultipleLookupValueFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MultipleLookupValue

    name = factory.Sequence(lambda n: 'multiplelookupvalue %s' % n)
    symbol = get_image(file_name='test_lookup_value_symbol.png')
    field = factory.SubFactory(MultipleLookupFieldFactory)
    status = 'active'



#!/usr/bin/env python

"""GeoKey setup."""

from os.path import join
from setuptools import setup, find_packages

from geokey.version import get_version


name = 'geokey'
version = get_version()
repository = join('https://github.com/ExCiteS', name)


def get_install_requires():
    """
    Get requirements (ignore links, exclude comments).

    Returns
    -------
    list
        Requirements for GeoKey.
    """
    requirements = list()
    for line in open('requirements.txt').readlines():
        if line.startswith('#') or line.startswith('git+https') or line == '':
            continue
        requirements.append(line.rstrip())
    return requirements


setup(
    name=name,
    version=version,
    description='Platform for participatory mapping',
    url='http://geokey.org.uk',
    download_url=join(repository, 'tarball', version),
    author='ExCiteS',
    author_email='excites@ucl.ac.uk',
    license='Apache 2.0',
    packages=find_packages(),
    include_package_data=True,
    install_requires=get_install_requires(),
)



"""Base for views of contributions."""

from geokey.projects.models import Project


class SingleAllContribution(object):
    """Base class for a single contribution of all contributions."""

    def get_contribution(self, user, project_id, contribution_id):
        """
        Get a single contribution.

        Parameters
        ----------
        user : geokey.users.models.User
            User requesting the request.
        project_id : int
            Identifies the project in the database.
        contribution_id : int
            Identifies the contribution in the database.

        Returns
        -------
        geokey.contributions.models.Observation
            Contribution with the required ID.
        """
        project = Project.objects.get_single(user, project_id)

        if project.can_moderate(user):
            return project\
                .get_all_contributions(user)\
                .for_moderator(user)\
                .select_related('location', 'project')\
                .prefetch_related('comments')\
                .get(pk=contribution_id)
        else:
            return project\
                .get_all_contributions(user)\
                .for_viewer(user)\
                .select_related('location', 'project')\
                .prefetch_related('comments')\
                .get(pk=contribution_id)



"""Tests for logger: model Location."""

from django.test import TestCase

from geokey.core.models import LoggerHistory
from geokey.users.tests.model_factories import UserFactory
from geokey.contributions.tests.model_factories import LocationFactory


class LogLocationFieldTest(TestCase):
    """Test model Location."""

    def setUp(self):
        """Set up test."""
        self.user = UserFactory.create()
        self.location = LocationFactory.create(**{
            'creator': self.user})

    def test_log_create(self):
        """Test when location gets created."""
        log_count_init = LoggerHistory.objects.count()
        location = LocationFactory.create(**{
            'creator': self.user})

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, None)
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, None)
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(location.id),
            'name': location.name})
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'created',
            'class': 'Location'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_delete(self):
        """Test when location gets deleted."""
        location_id = self.location.id
        location_name = self.location.name
        log_count_init = LoggerHistory.objects.count()
        self.location.delete()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, None)
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, None)
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(location_id),
            'name': location_name})
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'deleted',
            'class': 'Location'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_update_name(self):
        """Test when name changes."""
        log_count_init = LoggerHistory.objects.count()
        self.location.name = '%s UPDATED' % self.location.name
        self.location.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, None)
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, None)
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Location',
            'field': 'name'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_update_geometry(self):
        """Test when geometry changes."""
        log_count_init = LoggerHistory.objects.count()
        self.location.geometry = 'POINT(-0.134140712210241 51.12547879755655)'
        self.location.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, None)
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, None)
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Location',
            'field': 'geometry'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)



"""Tests for logger: model Observation."""

from django.test import TestCase

from geokey.core.models import LoggerHistory
from geokey.users.tests.model_factories import UserFactory
from geokey.projects.tests.model_factories import ProjectFactory
from geokey.categories.tests.model_factories import CategoryFactory
from geokey.contributions.tests.model_factories import (
    LocationFactory,
    ObservationFactory,
)


class LogObservationTest(TestCase):
    """Test model Observation."""

    def setUp(self):
        """Set up test."""
        self.user = UserFactory.create()
        self.project = ProjectFactory.create(**{
            'creator': self.user})
        self.category = CategoryFactory.create(**{
            'creator': self.user,
            'project': self.project})
        self.location = LocationFactory.create(**{
            'creator': self.user})
        self.observation = ObservationFactory.create(**{
            'creator': self.user,
            'location': self.location,
            'project': self.project,
            'category': self.category})

    def test_log_create(self):
        """Test when observation gets created."""
        log_count_init = LoggerHistory.objects.count()
        observation = ObservationFactory.create(**{
            'creator': self.user,
            'location': self.location,
            'project': self.project,
            'category': self.category})

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(observation.id)})
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'created',
            'class': 'Observation',
            'field': 'status',
            'value': observation.status})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_create_draft(self):
        """Test when observation gets created as a draft."""
        log_count_init = LoggerHistory.objects.count()
        ObservationFactory.create(**{
            'status': 'draft',
            'creator': self.user,
            'location': self.location,
            'project': self.project,
            'category': self.category})

        self.assertEqual(LoggerHistory.objects.count(), log_count_init)

    def test_log_create_from_draft_status(self):
        """Test when observation gets created from being a draft."""
        self.observation.status = 'draft'
        self.observation.save()
        log_count_init = LoggerHistory.objects.count()

        original_status = self.observation.status
        self.observation.status = 'pending'
        self.observation.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(self.observation.id)})
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'created',
            'class': 'Observation',
            'field': 'status',
            'value': self.observation.status})
        self.assertEqual(log_count, log_count_init + 1)
        history = self.observation.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.observation.id)
        self.assertEqual(history.status, original_status)

    def test_log_delete(self):
        """Test when observation gets deleted."""
        observation_id = self.observation.id
        log_count_init = LoggerHistory.objects.count()
        self.observation.delete()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(observation_id)})
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'deleted',
            'class': 'Observation',
            'field': 'status',
            'value': 'deleted'})
        self.assertEqual(log_count, log_count_init + 1)
        history = self.observation.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.observation.id)

    def test_log_update_status(self):
        """Test when observation status changes."""
        log_count_init = LoggerHistory.objects.count()

        original_status = self.observation.status
        self.observation.status = 'review'
        self.observation.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(self.observation.id)})
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Observation',
            'field': 'status',
            'value': 'review'})
        self.assertEqual(log_count, log_count_init + 1)
        history = self.observation.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.observation.id)
        self.assertEqual(history.status, original_status)

        original_status = self.observation.status
        self.observation.status = 'pending'
        self.observation.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(self.observation.id)})
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Observation',
            'field': 'status',
            'value': 'pending'})
        self.assertEqual(log_count, log_count_init + 2)
        history = self.observation.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.observation.id)
        self.assertEqual(history.status, original_status)

        original_status = self.observation.status
        self.observation.status = 'active'
        self.observation.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(self.observation.id)})
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Observation',
            'field': 'status',
            'value': 'active'})
        self.assertEqual(log_count, log_count_init + 3)
        history = self.observation.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.observation.id)
        self.assertEqual(history.status, original_status)

    def test_log_update_properties(self):
        """Test when observation properties changes."""
        log_count_init = LoggerHistory.objects.count()

        original_properties = self.observation.properties
        self.observation.properties = {'field': 'value'}
        self.observation.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertEqual(log_count, log_count_init + 1)

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(self.observation.id)})
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Observation',
            'field': 'properties'})
        self.assertEqual(log_count, log_count_init + 1)
        history = self.observation.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.observation.id)
        self.assertEqual(history.properties, original_properties)

    def test_log_update_properties_when_draft(self):
        """Test when observation properties changes, but status is "draft"."""
        self.observation.status = 'draft'
        self.observation.save()
        log_count_init = LoggerHistory.objects.count()

        self.observation.properties
        self.observation.properties = {'field': 'value'}
        self.observation.save()

        self.assertEqual(LoggerHistory.objects.count(), log_count_init)



"""Views for projects."""

from django.db import IntegrityError
from django.views.generic import CreateView, TemplateView
from django.shortcuts import redirect
from django.core.urlresolvers import reverse
from django.core.exceptions import PermissionDenied
from django.contrib import messages
from django.utils.html import strip_tags
from django.utils.safestring import mark_safe
from django.contrib.gis.geos import GEOSGeometry

from rest_framework import status
from rest_framework.views import APIView
from rest_framework.response import Response

from braces.views import LoginRequiredMixin

from geokey.core.decorators import (
    handle_exceptions_for_ajax,
    handle_exceptions_for_admin
)

from geokey.users.serializers import UserSerializer
from geokey.users.models import User
from geokey.categories.models import Category
from geokey.contributions.models import Comment, MediaFile

from .base import STATUS
from .models import Project, Admins
from .forms import ProjectCreateForm
from .serializers import ProjectSerializer


class ProjectContext(object):

    @handle_exceptions_for_admin
    def get_context_data(self, project_id, *args, **kwargs):
        """
        Returns the context containing the project instance.

        Parameters
        ----------
        project_id : int
            Identifies the project in the database

        Returns
        -------
        dict
            Context
        """

        project = Project.objects.as_admin(self.request.user, project_id)

        return super(ProjectContext, self).get_context_data(
            project=project,
            *args,
            **kwargs
        )


# ############################################################################
#
# Administration views
#
# ############################################################################

class ProjectCreate(LoginRequiredMixin, CreateView):

    """
    Displays the create project page.
    """
    form_class = ProjectCreateForm
    template_name = 'projects/project_create.html'

    def form_valid(self, form):
        """
        Creates the project and redirects to the project overview page.

        Parameters
        ----------
        form : geokey.projects.forms.ProjectCreateForm
            Represents the user input
        """

        data = form.cleaned_data

        project = Project.create(
            strip_tags(data.get('name')),
            strip_tags(data.get('description')),
            data.get('isprivate'),
            False,  # Project is never locked when creating it
            data.get('everyone_contributes'),
            self.request.user
        )

        add_another_url = reverse(
            'admin:project_create'
        )

        messages.success(
            self.request,
            mark_safe('The project has been created. <a href="%s">Add '
                      'another project.</a>' % add_another_url)
        )

        return redirect('admin:project_overview', project_id=project.id)


class ProjectsInvolved(LoginRequiredMixin, TemplateView):

    """
    Displays a list of all projects the user is involved in
    """
    template_name = 'projects/projects_involved.html'

    def get_context_data(self):
        """
        Returns the context to render the view. Overwrites the method to add
        the list of projects to the context.

        Returns
        -------
        dict
            Context
        """

        projects = Project.objects.get_list(self.request.user).exclude(
            admins=self.request.user)
        projects_list = []

        for project in projects:
            projects_list.append({
                'name': project.name,
                'role': project.get_role(self.request.user),
                'contributions': project.observations.filter(
                    creator=self.request.user).count(),
            })

        return {'projects': projects_list}


class ProjectOverview(LoginRequiredMixin, ProjectContext, TemplateView):
    """Project overview page."""

    template_name = 'projects/project_overview.html'

    def get_context_data(self, project_id):
        """
        Return the context to render the view.

        Overwrite the method to add the project, number of contributions,
        comments and media files in total.

        Parameters
        ----------
        project_id : int
            Identifies the project in the database.

        Returns
        -------
        dict
            Context
        """
        context = super(ProjectOverview, self).get_context_data(project_id)
        project = context.get('project')

        if project:
            contributions = project.observations.all()
            project.contributions_count = len(contributions)
            project.comments_count = Comment.objects.filter(
                commentto=contributions).count()
            project.media_count = MediaFile.objects.filter(
                contribution=contributions).count()

        return context


class ProjectGeographicExtent(LoginRequiredMixin, ProjectContext,
                              TemplateView):

    """
    Displays the page to edit the geograhic extent of the project.
    """
    template_name = 'projects/project_geographic_extent.html'

    def post(self, request, project_id):
        """
        Adds or updates the geographic extent of the project.

        Parameters
        ----------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database

        Returns
        -------
        django.http.HttpResponse
            Rendered template
        """

        data = request.POST
        context = self.get_context_data(project_id)
        project = context.get('project', None)
        geometry = data.get('geometry')

        if project is not None:
            if not project.islocked:
                if geometry is not None and len(geometry) > 0:
                    project.geographic_extent = GEOSGeometry(geometry)
                else:
                    project.geographic_extent = None

                project.save()
                context['project'] = project

                messages.success(
                    self.request,
                    'The geographic extent has been updated.'
                )
            else:
                messages.error(
                    self.request,
                    'The project is locked. Its structure cannot be edited.'
                )

        return self.render_to_response(context)


class ProjectSettings(LoginRequiredMixin, ProjectContext, TemplateView):

    """
    Displays the project settings page.
    """
    model = Project
    template_name = 'projects/project_settings.html'

    def get_context_data(self, project_id):
        """
        Returns the context to render the view. Overwrites the method to add
        the project and status types to the context.

        Parameters
        ----------
        project_id : int
            Identifies the project in the database

        Returns
        -------
        dict
            Context
        """

        return super(ProjectSettings, self).get_context_data(
            project_id,
            status_types=STATUS
        )

    def post(self, request, project_id):
        """
        Updates the project settings.

        Parameter
        ---------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database

        Returns
        -------
        django.http.HttpResponse
            Rendered template
        """

        context = self.get_context_data(project_id)
        project = context.get('project', None)

        if project is not None:
            data = request.POST

            project.name = strip_tags(data.get('name'))
            project.description = strip_tags(data.get('description'))
            project.everyone_contributes = data.get('everyone_contributes')
            project.save()
            context['project'] = project

            messages.success(self.request, 'The project has been updated.')

        return self.render_to_response(context)


class ProjectDelete(LoginRequiredMixin, ProjectContext, TemplateView):

    """
    Deletes the project.
    """
    template_name = 'base.html'

    def get(self, request, project_id):
        """
        Deletes the project.

        Parameters
        ----------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database

        Returns
        -------
        django.http.HttpResponseRedirect
            Redirecting to the dashboard
        django.http.HttpResponse
            If user is not administrator of the project, the error message is
            rendered
        """

        context = self.get_context_data(project_id)
        project = context.get('project', None)

        if project is not None:
            if not project.islocked:
                project.delete()
                messages.success(self.request, 'The project has been deleted.')
                return redirect('admin:dashboard')
            else:
                messages.error(
                    self.request,
                    'The project is locked. It cannot be deleted.'
                )
                return redirect(
                    'admin:project_settings', project_id=project.id)

        return redirect('admin:dashboard')


# ############################################################################
#
# AJAX API views
#
# ############################################################################

class ProjectUpdate(APIView):

    """
    AJAX Endpoint for a project update.
    /ajax/projects/:project_id
    """

    @handle_exceptions_for_ajax
    def put(self, request, project_id):
        """
        Updates a project.

        Parameter
        ---------
        request : rest_framework.request.Request
            Object representing the request.
        project_id : int
            identifies the project in the database

        Returns
        -------
        rest_framework.response.Response
            Response containing the serialised project or an error message
        """

        project = Project.objects.as_admin(request.user, project_id)

        serializer = ProjectSerializer(
            project, data=request.data, partial=True,
            fields=(
                'id', 'name', 'description', 'status', 'isprivate', 'islocked',
                'everyone_contributes'
            )
        )

        if serializer.is_valid():
            serializer.save()
            data = serializer.data
            data['everyone_contributes'] = project.everyone_contributes
            return Response(data)

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class ProjectAdmins(APIView):

    """
    AJAX Endpoint for project administrators.
    /ajax/projects/:project_id/admins
    """

    @handle_exceptions_for_ajax
    def post(self, request, project_id):
        """
        Adds a user to the admin group.

        Parameter
        ---------
        request : rest_framework.request.Request
            Object representing the request.
        project_id : int
            identifies the project in the database

        Returns
        -------
        rest_framework.response.Response
            Response containing the serialised list of admins or an error
            message.
        """

        project = Project.objects.as_admin(request.user, project_id)
        user = User.objects.get(pk=request.data.get('user_id'))

        try:
            Admins.objects.create(project=project, user=user)
        except IntegrityError:
            return Response(
                'The user is already an administrator of this project.',
                status=status.HTTP_400_BAD_REQUEST
            )

        refreshed_admins = Project.objects.get(pk=project_id).admins.all()
        serializer = UserSerializer(refreshed_admins, many=True)

        return Response(
            {'users': serializer.data},
            status=status.HTTP_201_CREATED
        )


class ProjectAdminsUser(APIView):

    """
    AJAX Endpoint for a single project administrator.
    /ajax/projects/:project_id/admins
    """

    @handle_exceptions_for_ajax
    def delete(self, request, project_id, user_id):
        """
        Removes a user from the user group.

        Parameter
        ---------
        request : rest_framework.request.Request
            Object representing the request.
        project_id : int
            identifies the project in the database

        Returns
        -------
        rest_framework.response.Response
            Empty response if successful or response containing an error
            message.
        """

        project = Project.objects.as_admin(request.user, project_id)
        user = project.admins.get(pk=user_id)
        Admins.objects.get(project=project, user=user).delete()

        return Response(status=status.HTTP_204_NO_CONTENT)


class CategoriesReorderView(APIView):

    """
    AJAX Endpoint to re-order categories in a project.
    /ajax/projects/:project_id/cotegories/re-order
    """

    @handle_exceptions_for_ajax
    def post(self, request, project_id):
        """
        Reorders the cateories in the project.

        Parameter
        ---------
        request : rest_framework.request.Request
            Object representing the request.
        project_id : int
            identifies the project in the database

        Returns
        -------
        rest_framework.response.Response
            Contains the serialised project or an error message
        """

        project = Project.objects.as_admin(request.user, project_id)

        try:
            project.reorder_categories(request.data.get('order'))

            serializer = ProjectSerializer(
                project,
                fields=(
                    'id', 'name', 'description', 'status', 'isprivate',
                    'islocked', 'everyone_contributes'
                )
            )
            return Response(serializer.data)
        except Category.DoesNotExist:
            return Response(
                {'error': 'One or more categories ids where not found.'},
                status=status.HTTP_400_BAD_REQUEST
            )


# ############################################################################
#
# PUBLIC API
#
# ############################################################################

class Projects(APIView):
    """Public API for all projects."""

    @handle_exceptions_for_ajax
    def get(self, request):
        """
        Handle GET request.

        Return a list a all projects accessible to the user.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.

        Returns
        -------
        rest_framework.response.Response
            Contains serialized list of projects.
        """
        user = request.user
        projects = Project.objects.get_list(user).filter(status='active')
        serializer = ProjectSerializer(
            projects,
            many=True,
            context={'user': user},
            fields=('id', 'name', 'description', 'user_info')
        )
        return Response(serializer.data)


class SingleProject(APIView):
    """Public API for a single project."""

    @handle_exceptions_for_ajax
    def get(self, request, project_id):
        """
        Handle GET request.

        Return a single project.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        project_id : int
            Identifies the project in the database.

        Returns
        -------
        rest_framework.response.Response
            Contains the serialized project.

        Raises
        ------
        PermissionDenied
            When the project is inactive (handled in the
            handle_exceptions_for_ajax decorator).
        """
        user = request.user
        project = Project.objects.get_single(user, project_id)

        if project.status == 'active':
            serializer = ProjectSerializer(
                project,
                context={'user': user}
            )
            return Response(serializer.data)

        raise PermissionDenied('The project is inactive and therefore '
                               'not accessable through the public API.')



"""Tests for views of superuser tools."""

from collections import OrderedDict
from importlib import import_module

from django.test import TestCase
from django.conf import settings
from django.apps import apps
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect, HttpRequest, QueryDict
from django.template.loader import render_to_string
from django.contrib.auth.models import AnonymousUser
from django.contrib.sites.models import Site
from django.contrib.sites.shortcuts import get_current_site
from django.contrib.messages import get_messages
from django.contrib.messages.storage.fallback import FallbackStorage

from allauth.account.models import EmailAddress
from allauth.socialaccount import providers
from allauth.socialaccount.models import SocialApp
from rest_framework.test import APIRequestFactory, force_authenticate

from geokey import version
from geokey.core.tests.helpers import render_helpers
from geokey.users.models import User
from geokey.users.tests.model_factories import UserFactory
from geokey.projects.tests.model_factories import ProjectFactory
from geokey.categories.tests.model_factories import CategoryFactory
from geokey.contributions.tests.model_factories import (
    ObservationFactory,
    CommentFactory
)

from geokey.superusertools.views import (
    ManageSuperusers,
    ManageInactiveUsers,
    ManageProjects,
    PlatformSettings,
    ProviderList,
    ProviderOverview,
    SuperusersAjaxView,
    SingleSuperuserAjaxView
)


# #############################################################################
#
# ADMIN VIEWS
#
# #############################################################################

class ManageSuperusersTest(TestCase):
    """Test manage superusers page."""

    def setUp(self):
        """Set up test."""
        self.url = reverse('admin:superusertools_manage_superusers')
        self.request = APIRequestFactory().get(self.url)

    def test_get_context_data(self):
        """Test getting context data."""
        UserFactory.create_batch(2, **{'is_superuser': True})
        UserFactory.create_batch(2, **{'is_superuser': False})
        view = ManageSuperusers()
        context = view.get_context_data()

        self.assertEqual(len(context.get('superusers')), 2)

    def test_get_with_anonymous(self):
        """Test GET with anonymous user."""
        view = ManageSuperusers.as_view()
        self.request.user = AnonymousUser()
        response = view(self.request)

        self.assertTrue(isinstance(response, HttpResponseRedirect))

    def test_get_with_user(self):
        """Test GET with user."""
        view = ManageSuperusers.as_view()
        self.request.user = UserFactory.create(**{'is_superuser': False})
        response = view(self.request).render()

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'No rights to access superuser tools.'
        )

    def test_get_with_superuser(self):
        """Test GET with superuser."""
        view = ManageSuperusers.as_view()
        self.request.user = UserFactory.create(**{'is_superuser': True})
        response = view(self.request).render()

        self.assertEqual(response.status_code, 200)


class ManageInactiveUsersTest(TestCase):
    """Test manage inactivate users page."""

    def setUp(self):
        """Set up test."""
        self.view = ManageInactiveUsers.as_view()
        self.request = HttpRequest()

        setattr(self.request, 'session', 'session')
        messages = FallbackStorage(self.request)
        setattr(self.request, '_messages', messages)

    def create_inactive(self):
        """Create 3 inactive users."""
        self.inactive_1 = UserFactory.create(**{'is_active': False})
        EmailAddress(
            user=self.inactive_1,
            email=self.inactive_1.email,
            verified=False
        ).save()
        self.inactive_2 = UserFactory.create(**{'is_active': False})
        EmailAddress(
            user=self.inactive_2,
            email=self.inactive_2.email,
            verified=False
        ).save()
        self.inactive_3 = UserFactory.create(**{'is_active': False})
        EmailAddress(
            user=self.inactive_3,
            email=self.inactive_3.email,
            verified=False
        ).save()

    def test_get_with_anonymous(self):
        """Test GET with anonymous user."""
        self.request.method = 'GET'
        self.request.user = AnonymousUser()
        response = self.view(self.request)

        self.assertEqual(response.status_code, 302)
        self.assertIn('/admin/account/login/', response['location'])

    def test_get_with_user(self):
        """Test GET with user."""
        user = UserFactory.create(**{'is_superuser': False})
        self.request.method = 'GET'
        self.request.user = user
        response = self.view(self.request).render()

        rendered = render_to_string(
            'superusertools/manage_inactive_users.html',
            {
                'GEOKEY_VERSION': version.get_version(),
                'PLATFORM_NAME': get_current_site(self.request).name,
                'user': user,
                'error': 'Permission denied.',
                'error_description': 'No rights to access superuser tools.',
            }
        )

        self.assertEqual(response.status_code, 200)
        response = render_helpers.remove_csrf(response.content.decode('utf-8'))
        self.assertEqual(response, rendered)

    def test_get_with_superuser(self):
        """Test GET with superuser."""
        self.create_inactive()
        user = UserFactory.create(**{'is_superuser': True})
        self.request.method = 'GET'
        self.request.user = user
        response = self.view(self.request).render()

        rendered = render_to_string(
            'superusertools/manage_inactive_users.html',
            {
                'GEOKEY_VERSION': version.get_version(),
                'PLATFORM_NAME': get_current_site(self.request).name,
                'user': user,
                'inactive_users': [
                    self.inactive_1,
                    self.inactive_2,
                    self.inactive_3
                ]
            }
        )

        self.assertEqual(response.status_code, 200)
        response = render_helpers.remove_csrf(response.content.decode('utf-8'))
        self.assertEqual(response, rendered)

    def test_post_with_anonymous(self):
        """Test POST with anonymous user."""
        self.create_inactive()
        self.request.method = 'POST'
        self.request.POST = QueryDict(
            'activate_users=%s&activate_users=%s' % (
                self.inactive_1.id, self.inactive_2.id
            )
        )
        self.request.user = AnonymousUser()
        response = self.view(self.request)

        self.assertEqual(response.status_code, 302)
        self.assertIn('/admin/account/login/', response['location'])
        self.assertEqual(User.objects.filter(is_active=False).count(), 3)
        self.assertEqual(len(EmailAddress.objects.filter(verified=False)), 3)

    def test_post_with_user(self):
        """Test POST with user."""
        self.create_inactive()
        user = UserFactory.create()
        self.request.method = 'POST'
        self.request.POST = QueryDict(
            'activate_users=%s&activate_users=%s' % (
                self.inactive_1.id, self.inactive_2.id
            )
        )
        self.request.user = user
        response = self.view(self.request).render()

        rendered = render_to_string(
            'superusertools/manage_inactive_users.html',
            {
                'GEOKEY_VERSION': version.get_version(),
                'PLATFORM_NAME': get_current_site(self.request).name,
                'user': user,
                'error': 'Permission denied.',
                'error_description': 'No rights to access superuser tools.'
            }
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode('utf-8'), rendered)
        response = render_helpers.remove_csrf(response.content.decode('utf-8'))
        self.assertEqual(User.objects.filter(is_active=False).count(), 3)
        self.assertEqual(len(EmailAddress.objects.filter(verified=False)), 3)

    def test_post_with_superuser(self):
        """Test POST with superuser."""
        self.create_inactive()
        user = UserFactory.create(**{'is_superuser': True})
        self.request.method = 'POST'
        self.request.POST = QueryDict(
            'activate_users=%s&activate_users=%s' % (
                self.inactive_1.id, self.inactive_2.id
            )
        )
        self.request.user = user
        response = self.view(self.request).render()

        rendered = render_to_string(
            'superusertools/manage_inactive_users.html',
            {
                'GEOKEY_VERSION': version.get_version(),
                'PLATFORM_NAME': get_current_site(self.request).name,
                'user': user,
                'messages': get_messages(self.request),
                'inactive_users': [self.inactive_3]
            }
        )

        self.assertEqual(response.status_code, 200)
        response = render_helpers.remove_csrf(response.content.decode('utf-8'))
        self.assertEqual(response, rendered)
        self.assertEqual(User.objects.filter(is_active=False).count(), 1)
        self.assertEqual(len(EmailAddress.objects.filter(verified=False)), 1)


class ManageProjectsTest(TestCase):
    """Test manage projects page."""

    def setUp(self):
        """Set up test."""
        self.url = reverse('admin:superusertools_manage_projects')
        self.request = APIRequestFactory().get(self.url)

    def test_get_context_data(self):
        """Test getting context data."""
        user = UserFactory.create(**{'is_superuser': True})

        # 1 contribution, 1 comment
        project_1 = ProjectFactory.create()
        category_1 = CategoryFactory.create(project=project_1)
        contribution_1 = ObservationFactory.create(
            project=project_1,
            category=category_1)
        CommentFactory.create(commentto=contribution_1)

        # 2 contributions (1 deleted), 2 comments
        project_2 = ProjectFactory.create(add_admins=[user])
        category_2 = CategoryFactory.create(project=project_2)
        contribution_2 = ObservationFactory.create(
            project=project_2,
            category=category_2)
        CommentFactory.create(commentto=contribution_2)
        contribution_3 = ObservationFactory.create(
            project=project_2,
            category=category_2)
        CommentFactory.create(commentto=contribution_3)
        contribution_3.delete()

        # 2 contributions (1 deleted), 3 comments (1 deleted)
        project_3 = ProjectFactory.create(add_moderators=[user])
        category_3 = CategoryFactory.create(project=project_3)
        contribution_4 = ObservationFactory.create(
            project=project_3,
            category=category_3)
        CommentFactory.create(commentto=contribution_4)
        comment_to_delete = CommentFactory.create(commentto=contribution_4)
        comment_to_delete.delete()
        contribution_5 = ObservationFactory.create(
            project=project_3,
            category=category_3)
        CommentFactory.create(commentto=contribution_5)
        contribution_5.delete()

        # 1 contribution, 2 comments (1 deleted)
        project_4 = ProjectFactory.create(add_contributors=[user])
        category_4 = CategoryFactory.create(project=project_4)
        contribution_6 = ObservationFactory.create(
            project=project_4,
            category=category_4)
        CommentFactory.create(commentto=contribution_6)
        comment_to_delete = CommentFactory.create(commentto=contribution_6)
        comment_to_delete.delete()

        view = ManageProjects()
        context = view.get_context_data()
        self.assertEqual(len(context.get('projects')), 4)

        for project in context.get('projects'):
            self.assertEqual(project.contributions_count, 1)
            self.assertEqual(project.comments_count, 1)
            self.assertEqual(project.media_count, 0)

    def test_get_with_anonymous(self):
        """Test GET with anonymous user."""
        view = ManageProjects.as_view()
        self.request.user = AnonymousUser()
        response = view(self.request)

        self.assertTrue(isinstance(response, HttpResponseRedirect))

    def test_get_with_user(self):
        """Test GET with user."""
        view = ManageProjects.as_view()
        self.request.user = UserFactory.create(**{'is_superuser': False})
        response = view(self.request).render()

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'No rights to access superuser tools.'
        )

    def test_get_with_superuser(self):
        """Test GET with superuser."""
        view = ManageProjects.as_view()
        self.request.user = UserFactory.create(**{'is_superuser': True})
        response = view(self.request).render()

        self.assertEqual(response.status_code, 200)


class PlatformSettingsTest(TestCase):
    """Test platform settings page."""

    def setUp(self):
        """Set up test."""
        self.url = reverse('admin:superusertools_platform_settings')

    def test_get_context_data(self):
        """Test getting context data."""
        view = PlatformSettings()
        view.request = APIRequestFactory().get(self.url)
        context = view.get_context_data()

        self.assertIsNotNone(context.get('site'))

    def test_get_with_anonymous(self):
        """Test GET with anonymous user."""
        view = PlatformSettings.as_view()
        request = APIRequestFactory().get(self.url)
        request.user = AnonymousUser()
        view.request = request
        response = view(request)

        self.assertTrue(isinstance(response, HttpResponseRedirect))

    def test_get_with_user(self):
        """Test GET with user."""
        view = PlatformSettings.as_view()
        request = APIRequestFactory().get(self.url)
        request.user = UserFactory.create(**{'is_superuser': False})
        view.request = request
        response = view(request).render()

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'No rights to access superuser tools.'
        )

    def test_get_with_superuser(self):
        """Test GET with superuser."""
        view = PlatformSettings.as_view()
        request = APIRequestFactory().get(self.url)
        request.user = UserFactory.create(**{'is_superuser': True})
        response = view(request).render()

        self.assertEqual(response.status_code, 200)
        self.assertNotContains(
            response,
            'No rights to access superuser tools.'
        )

    def test_post_with_anonymous(self):
        """Test POST with anonymous user."""
        data = {
            'name': 'New Name',
            'domain': 'http://new-domain.org.uk'
        }
        view = PlatformSettings.as_view()
        request = APIRequestFactory().post(self.url, data)
        request.user = AnonymousUser()
        view.request = request
        response = view(request)

        self.assertTrue(isinstance(response, HttpResponseRedirect))

        reference = Site.objects.latest('id')
        self.assertNotEqual(reference.name, data.get('name'))
        self.assertNotEqual(reference.domain, data.get('domain'))

    def test_post_with_user(self):
        """Test POST with user."""
        data = {
            'name': 'New Name',
            'domain': 'http://new-domain.org.uk'
        }
        view = PlatformSettings.as_view()
        request = APIRequestFactory().post(self.url, data)
        request.user = UserFactory.create(**{'is_superuser': False})
        view.request = request
        response = view(request).render()

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'No rights to access superuser tools.'
        )

        reference = Site.objects.latest('id')
        self.assertNotEqual(reference.name, data.get('name'))
        self.assertNotEqual(reference.domain, data.get('domain'))

    def test_post_with_superuser(self):
        """Test POST with superuser."""
        data = {
            'name': 'New Name',
            'domain': 'http://new-domain.org.uk'
        }
        view = PlatformSettings.as_view()
        request = APIRequestFactory().post(self.url, data)
        request.user = UserFactory.create(**{'is_superuser': True})

        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        setattr(request, '_messages', messages)

        response = view(request).render()
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Platform settings have been updated.')

        reference = Site.objects.latest('id')
        self.assertEqual(reference.name, data.get('name'))
        self.assertEqual(reference.domain, data.get('domain'))


class ProviderListTest(TestCase):
    """Test a list of all providers page."""

    def setUp(self):
        """Set up test."""
        self.url = reverse('admin:superusertools_provider_list')

    def test_get_context_data(self):
        """Test getting context data."""
        view = ProviderList()
        view.request = APIRequestFactory().get(self.url)
        context = view.get_context_data()

        self.assertIsNotNone(context.get('providers'))

    def test_get_with_anonymous(self):
        """Test GET with anonymous user."""
        view = ProviderList.as_view()
        request = APIRequestFactory().get(self.url)
        request.user = AnonymousUser()
        view.request = request
        response = view(request)

        self.assertTrue(isinstance(response, HttpResponseRedirect))

    def test_get_with_user(self):
        """Test GET with user."""
        view = ProviderList.as_view()
        request = APIRequestFactory().get(self.url)
        request.user = UserFactory.create(**{'is_superuser': False})
        view.request = request
        response = view(request).render()

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'No rights to access superuser tools.'
        )

    def test_get_with_superuser(self):
        """Test GET with superuser."""
        view = ProviderList.as_view()
        request = APIRequestFactory().get(self.url)
        request.user = UserFactory.create(**{'is_superuser': True})
        view.request = request
        response = view(request).render()

        self.assertEqual(response.status_code, 200)
        self.assertNotContains(
            response,
            'No rights to access superuser tools.'
        )


class ProviderOverviewTest(TestCase):
    """Test overview of a provider page."""

    def setUp(self):
        """Set up test."""
        google_provider = 'allauth.socialaccount.providers.google'
        if google_provider not in settings.INSTALLED_APPS:
            settings.INSTALLED_APPS += (google_provider,)
            apps.app_configs = OrderedDict()
            apps.ready = False
            apps.populate(settings.INSTALLED_APPS)
            module = import_module(google_provider + '.provider')
            for cls in getattr(module, 'provider_classes', []):
                providers.registry.register(cls)

        self.url = reverse(
            'admin:superusertools_provider_overview',
            kwargs={'provider_id': 'google'}
        )

    def test_get_context_data_when_social_app_doesnt_exist(self):
        """Test getting context data when social app doesn't exist."""
        view = ProviderOverview()
        view.request = APIRequestFactory().get(self.url)

        social_app = SocialApp.objects.create(
            provider='facebook',
            name='Facebook',
            client_id='xxxxxxxxxxxxxxxxxx',
            secret='xxxxxxxxxxxxxxxxxx',
            key=''
        )
        social_app.sites.add(get_current_site(view.request))

        context = view.get_context_data('google')

        self.assertIsNotNone(context.get('provider'))
        self.assertIsNone(context.get('social_app'))

    def test_get_context_data_when_social_app_exists(self):
        """Test getting context data when social app exists."""
        view = ProviderOverview()
        view.request = APIRequestFactory().get(self.url)

        social_app = SocialApp.objects.create(
            provider='google',
            name='Google',
            client_id='xxxxxxxxxxxxxxxxxx',
            secret='xxxxxxxxxxxxxxxxxx',
            key=''
        )
        social_app.sites.add(get_current_site(view.request))

        context = view.get_context_data('google')

        self.assertIsNotNone(context.get('provider'))
        self.assertEqual(context.get('social_app'), social_app)

    def test_get_context_data_when_social_app_exists_on_wrong_site(self):
        """Test getting context data when social app exists on wrong site."""
        view = ProviderOverview()
        view.request = APIRequestFactory().get(self.url)

        SocialApp.objects.create(
            provider='google',
            name='Google',
            client_id='xxxxxxxxxxxxxxxxxx',
            secret='xxxxxxxxxxxxxxxxxx',
            key=''
        )

        context = view.get_context_data('google')

        self.assertIsNotNone(context.get('provider'))
        self.assertIsNone(context.get('social_app'))

    def test_get_context_data_when_provider_doesnt_exist(self):
        """Test getting context data when provider doesn't exist."""
        url = reverse(
            'admin:superusertools_provider_overview',
            kwargs={'provider_id': 'myawesomeprovider'}
        )
        view = ProviderOverview.as_view()
        request = APIRequestFactory().get(url)
        request.user = UserFactory.create(**{'is_superuser': True})
        view.request = request
        response = view(request, provider_id='myawesomeprovider').render()

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'Provider not found.'
        )

    def test_get_with_anonymous(self):
        """Test GET with anonymous user."""
        view = ProviderOverview.as_view()
        request = APIRequestFactory().get(self.url)
        request.user = AnonymousUser()
        view.request = request
        response = view(request, provider_id='google')

        self.assertTrue(isinstance(response, HttpResponseRedirect))

    def test_get_with_user(self):
        """Test GET with user."""
        view = ProviderOverview.as_view()
        request = APIRequestFactory().get(self.url)
        request.user = UserFactory.create(**{'is_superuser': False})
        view.request = request
        response = view(request, provider_id='google').render()

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'No rights to access superuser tools.'
        )

    def test_get_with_superuser(self):
        """Test GET with superuser."""
        view = ProviderOverview.as_view()
        request = APIRequestFactory().get(self.url)
        request.user = UserFactory.create(**{'is_superuser': True})
        view.request = request
        response = view(request, provider_id='google').render()

        self.assertEqual(response.status_code, 200)
        self.assertNotContains(
            response,
            'No rights to access superuser tools.'
        )

    def test_post_with_anonymous(self):
        """Test POST with anonymous user."""
        data = {
            'client_id': 'xxxxxxxxxxxxxxxxxx',
            'secret': 'xxxxxxxxxxxxxxxxxx',
            'key': ''
        }
        view = ProviderOverview.as_view()
        request = APIRequestFactory().post(self.url, data)
        request.user = AnonymousUser()
        view.request = request
        response = view(request, provider_id='google')

        self.assertTrue(isinstance(response, HttpResponseRedirect))
        self.assertEqual(SocialApp.objects.count(), 0)

    def test_post_with_user(self):
        """Test POST with user."""
        data = {
            'client_id': 'xxxxxxxxxxxxxxxxxx',
            'secret': 'xxxxxxxxxxxxxxxxxx',
            'key': ''
        }
        view = ProviderOverview.as_view()
        request = APIRequestFactory().post(self.url, data)
        request.user = UserFactory.create(**{'is_superuser': False})
        view.request = request
        response = view(request, provider_id='google').render()

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'No rights to access superuser tools.'
        )
        self.assertEqual(SocialApp.objects.count(), 0)

    def test_post_with_superuser_when_activating(self):
        """Test POST with superuser."""
        data = {
            'client_id': 'xxxxxxxxxxxxxxxxxx',
            'secret': 'xxxxxxxxxxxxxxxxxx',
            'key': ''
        }
        view = ProviderOverview.as_view()
        request = APIRequestFactory().post(self.url, data)
        request.user = UserFactory.create(**{'is_superuser': True})

        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        setattr(request, '_messages', messages)

        response = view(request, provider_id='google').render()
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Provider has been activated.')

        reference = SocialApp.objects.latest('id')
        self.assertEqual(reference.client_id, data.get('client_id'))
        self.assertEqual(reference.secret, data.get('secret'))
        self.assertEqual(reference.key, data.get('key'))

    def test_post_with_superuser_when_updating(self):
        """Test POST with superuser."""
        data = {
            'client_id': 'xxxxxxxxxxxxxxxxxx',
            'secret': 'xxxxxxxxxxxxxxxxxx',
            'key': ''
        }
        view = ProviderOverview.as_view()
        request = APIRequestFactory().post(self.url, data)
        request.user = UserFactory.create(**{'is_superuser': True})

        social_app = SocialApp.objects.create(
            provider='google',
            name='Google',
            client_id='yyyyyyyyyyyyyyyyyy',
            secret='yyyyyyyyyyyyyyyyyy',
            key=''
        )
        social_app.sites.add(get_current_site(request))

        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        setattr(request, '_messages', messages)

        response = view(request, provider_id='google').render()
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Provider has been updated.')
        self.assertEqual(SocialApp.objects.latest('id'), social_app)

        reference = SocialApp.objects.get(pk=social_app.id)
        self.assertEqual(reference.client_id, data.get('client_id'))
        self.assertEqual(reference.secret, data.get('secret'))
        self.assertEqual(reference.key, data.get('key'))


# #############################################################################
#
# AJAX API
#
# #############################################################################

class SuperusersAjaxViewTest(TestCase):
    """Test Ajax API for all superusers."""

    def setUp(self):
        """Set up test."""
        self.factory = APIRequestFactory()
        self.view = SuperusersAjaxView.as_view()
        self.url = reverse('ajax:superusertools_superusers')

        self.user = UserFactory.create(**{'is_superuser': False})
        self.superuser = UserFactory.create(**{'is_superuser': True})
        self.user_to_add = UserFactory.create(**{'is_superuser': False})

    def test_post_with_anonymous_user(self):
        """Test POST with anonymous user."""
        request = self.factory.post(self.url, {'user_id': self.user_to_add.id})
        response = self.view(request).render()

        self.assertEqual(response.status_code, 403)
        self.assertEqual(User.objects.filter(is_superuser=True).count(), 1)

    def test_post_with_user(self):
        """Test POST with user."""
        request = self.factory.post(self.url, {'user_id': self.user_to_add.id})
        force_authenticate(request, user=self.user)
        response = self.view(request).render()

        self.assertEqual(response.status_code, 403)
        self.assertEqual(User.objects.filter(is_superuser=True).count(), 1)

    def test_post_with_superuser(self):
        """Test POST with superuser."""
        request = self.factory.post(self.url, {'user_id': self.user_to_add.id})
        force_authenticate(request, user=self.superuser)
        response = self.view(request).render()

        self.assertEqual(response.status_code, 201)
        self.assertEqual(User.objects.filter(is_superuser=True).count(), 2)

    def test_post_when_no_superuser(self):
        """Test POST with superuser, when user does not exist."""
        request = self.factory.post(self.url, {'user_id': 84774358734})
        force_authenticate(request, user=self.superuser)
        response = self.view(request).render()

        self.assertEqual(response.status_code, 404)
        self.assertEqual(User.objects.filter(is_superuser=True).count(), 1)


class SingleSuperuserAjaxViewTest(TestCase):
    """Test Ajax API for a single superuser."""

    def setUp(self):
        """Set up test."""
        self.factory = APIRequestFactory()
        self.view = SingleSuperuserAjaxView.as_view()

        self.user = UserFactory.create(**{'is_superuser': False})
        self.superuser = UserFactory.create(**{'is_superuser': True})
        self.user_to_remove = UserFactory.create(**{'is_superuser': True})

    def test_delete_with_anonymous_user(self):
        """Test DELETE with anonymous user."""
        request = self.factory.delete(
            reverse('ajax:superusertools_single_superuser', kwargs={
                'user_id': self.user_to_remove.id
            })
        )
        response = self.view(request, user_id=self.user_to_remove.id).render()

        self.assertEqual(response.status_code, 403)
        self.assertEqual(User.objects.filter(is_superuser=True).count(), 2)

    def test_delete_with_user(self):
        """Test DELETE with user."""
        request = self.factory.delete(
            reverse('ajax:superusertools_single_superuser', kwargs={
                'user_id': self.user_to_remove.id
            })
        )
        force_authenticate(request, user=self.user)
        response = self.view(request, user_id=self.user_to_remove.id).render()

        self.assertEqual(response.status_code, 403)
        self.assertEqual(User.objects.filter(is_superuser=True).count(), 2)

    def test_delete_with_superuser(self):
        """Test DELETE with superuser."""
        request = self.factory.delete(
            reverse('ajax:superusertools_single_superuser', kwargs={
                'user_id': self.user_to_remove.id
            })
        )
        force_authenticate(request, user=self.superuser)
        response = self.view(request, user_id=self.user_to_remove.id).render()

        self.assertEqual(response.status_code, 204)
        self.assertEqual(User.objects.filter(is_superuser=True).count(), 1)

    def test_delete_when_no_superuser(self):
        """Test DELETE with superuser, when superuser does not exist."""
        request = self.factory.delete(
            reverse('ajax:superusertools_single_superuser', kwargs={
                'user_id': 84774358734
            })
        )
        force_authenticate(request, user=self.superuser)
        response = self.view(request, user_id=84774358734).render()

        self.assertEqual(response.status_code, 404)
        self.assertEqual(User.objects.filter(is_superuser=True).count(), 2)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django_pgjson.fields
import django.db.models.deletion
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('projects', '0008_historicalproject'),
        ('users', '0007_auto_20151006_1110'),
    ]

    operations = [
        migrations.CreateModel(
            name='HistoricalUserGroup',
            fields=[
                ('id', models.IntegerField(verbose_name='ID', db_index=True, auto_created=True, blank=True)),
                ('name', models.CharField(max_length=100)),
                ('description', models.TextField(null=True, blank=True)),
                ('can_contribute', models.BooleanField(default=True)),
                ('can_moderate', models.BooleanField(default=False)),
                ('filters', django_pgjson.fields.JsonBField(null=True, blank=True)),
                ('where_clause', models.TextField(null=True, blank=True)),
                ('history_id', models.AutoField(serialize=False, primary_key=True)),
                ('history_date', models.DateTimeField()),
                ('history_type', models.CharField(max_length=1, choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')])),
                ('history_user', models.ForeignKey(related_name='+', on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, null=True)),
                ('project', models.ForeignKey(related_name='+', on_delete=django.db.models.deletion.DO_NOTHING, db_constraint=False, blank=True, to='projects.Project', null=True)),
            ],
            options={
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
                'verbose_name': 'historical user group',
            },
        ),
    ]



"""Admin URLs."""

from django.conf.urls import url
from django.core.urlresolvers import reverse_lazy
from django.views.generic.base import RedirectView

from geokey.projects import views as project_views
from geokey.categories import views as category_views
from geokey.users import views as user_views
from geokey.applications import views as app_views
from geokey.superusertools import views as superusertools
from geokey.subsets import views as subsets


urlpatterns = [
    url(r'^$', user_views.Index.as_view(), name='index'),
    url(r'^dashboard/$', user_views.Dashboard.as_view(), name='dashboard'),

    # ###########################
    # PROJECTS
    # ###########################
    url(
        r'^projects/new/$',
        project_views.ProjectCreate.as_view(),
        name='project_create'),
    url(
        r'^projects/involved/$',
        project_views.ProjectsInvolved.as_view(),
        name='projects_involved'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/$',
        project_views.ProjectOverview.as_view(),
        name='project_overview'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/geographicextent/$',
        project_views.ProjectGeographicExtent.as_view(),
        name='project_geographicextent'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/settings/$',
        project_views.ProjectSettings.as_view(),
        name='project_settings'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/delete/$',
        project_views.ProjectDelete.as_view(),
        name='project_delete'),


    # ###########################
    # USER GROUPS
    # ###########################
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'usergroups/$',
        user_views.UserGroupList.as_view(),
        name='usergroup_list'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'usergroups/new/$',
        user_views.UserGroupCreate.as_view(),
        name='usergroup_create'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'usergroups/(?P<usergroup_id>[0-9]+)/$',
        user_views.UserGroupOverview.as_view(),
        name='usergroup_overview'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'usergroups/administrators/$',
        user_views.AdministratorsOverview.as_view(),
        name='admins_overview'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'usergroups/(?P<usergroup_id>[0-9]+)/permissions/$',
        user_views.UserGroupPermissions.as_view(),
        name='usergroup_permissions'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'usergroups/(?P<usergroup_id>[0-9]+)/settings/$',
        user_views.UserGroupSettings.as_view(),
        name='usergroup_settings'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'usergroups/(?P<usergroup_id>[0-9]+)/delete/$',
        user_views.UserGroupDelete.as_view(),
        name='usergroup_delete'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'usergroups/(?P<usergroup_id>[0-9]+)/data/$',
        user_views.UserGroupData.as_view(),
        name='usergroup_data'),

    # ###########################
    # CATEGORIES & FIELDS
    # ###########################
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'categories/$',
        category_views.CategoryList.as_view(),
        name='category_list'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'categories/new/$',
        category_views.CategoryCreate.as_view(),
        name='category_create'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'categories/(?P<category_id>[0-9]+)/$',
        category_views.CategoryOverview.as_view(),
        name='category_overview'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'categories/(?P<category_id>[0-9]+)/display/$',
        category_views.CategoryDisplay.as_view(),
        name='category_display'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'categories/(?P<category_id>[0-9]+)/settings/$',
        category_views.CategorySettings.as_view(),
        name='category_settings'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'categories/(?P<category_id>[0-9]+)/delete/$',
        category_views.CategoryDelete.as_view(),
        name='category_delete'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'categories/(?P<category_id>[0-9]+)/'
        r'fields/new/$',
        category_views.FieldCreate.as_view(),
        name='category_field_create'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'categories/(?P<category_id>[0-9]+)/'
        r'fields/(?P<field_id>[0-9]+)/$',
        category_views.FieldSettings.as_view(),
        name='category_field_settings'),
    url(
        r'^projects/(?P<project_id>[0-9]+)/'
        r'categories/(?P<category_id>[0-9]+)/'
        r'fields/(?P<field_id>[0-9]+)/delete/$',
        category_views.FieldDelete.as_view(),
        name='category_field_delete'),

    # ###########################
    # SUBSETS
    # ###########################
    url(r'^projects/(?P<project_id>[0-9]+)/'
        r'subsets/$',
        subsets.SubsetList.as_view(),
        name='subset_list'),
    url(r'^projects/(?P<project_id>[0-9]+)/'
        r'subsets/new/$',
        subsets.SubsetCreate.as_view(),
        name='subset_create'),
    url(r'^projects/(?P<project_id>[0-9]+)/'
        r'subsets/(?P<subset_id>[0-9]+)/$',
        subsets.SubsetSettings.as_view(),
        name='subset_settings'),
    url(r'^projects/(?P<project_id>[0-9]+)/'
        r'subsets/(?P<subset_id>[0-9]+)/data/$',
        subsets.SubsetData.as_view(),
        name='subset_data'),
    url(r'^projects/(?P<project_id>[0-9]+)/'
        r'subsets/(?P<subset_id>[0-9]+)/delete/$',
        subsets.SubsetDelete.as_view(),
        name='subset_delete'),

    # ###########################
    # APPS
    # ###########################
    url(
        r'^apps/$',
        app_views.ApplicationOverview.as_view(),
        name='app_overview'),
    url(
        r'^apps/connected/$',
        app_views.ApplicationConnected.as_view(),
        name='app_connected'),
    url(
        r'^apps/register/$',
        app_views.ApplicationCreate.as_view(),
        name='app_register'),
    url(
        r'^apps/(?P<app_id>[0-9]+)/$',
        app_views.ApplicationSettings.as_view(),
        name='app_settings'),
    url(
        r'^apps/(?P<app_id>[0-9]+)/delete/$',
        app_views.ApplicationDelete.as_view(),
        name='app_delete'),
    url(
        r'^apps/(?P<app_id>[0-9]+)/disconnect/$',
        app_views.ApplicationDisconnect.as_view(),
        name='app_disconnect'),

    # ###########################
    # USERS
    # ###########################
    url(r'^profile/$',
        user_views.UserProfile.as_view(),
        name='userprofile'),
    url(r'^accounts/(?P<account_id>[0-9]+)/disconnect/$',
        user_views.AccountDisconnect.as_view(),
        name='account_disconnect'),

    # ###########################
    # SUPERUSER TOOLS
    # ###########################
    url(r'^superusertools/$',
        RedirectView.as_view(
            url=reverse_lazy('admin:superusertools_manage_superusers'),
            permanent=False),
        name='superusertools_index'),
    url(r'^superusertools/manage-superusers/$',
        superusertools.ManageSuperusers.as_view(),
        name='superusertools_manage_superusers'),
    url(r'^superusertools/manage-inactive-users/$',
        superusertools.ManageInactiveUsers.as_view(),
        name='superusertools_manage_inactive_users'),
    url(r'^superusertools/manage-projects/$',
        superusertools.ManageProjects.as_view(),
        name='superusertools_manage_projects'),
    url(r'^superuser-tools/platform-settings/$',
        superusertools.PlatformSettings.as_view(),
        name='superusertools_platform_settings'),
    url(r'^superuser-tools/providers/$',
        superusertools.ProviderList.as_view(),
        name='superusertools_provider_list'),
    url(r'^superuser-tools/providers/(?P<provider_id>[\w-]+)/$',
        superusertools.ProviderOverview.as_view(),
        name='superusertools_provider_overview'),
]



"""Core views."""

from rest_framework.views import APIView
from rest_framework.response import Response

from geokey.version import get_version
from geokey.extensions.base import extensions


# ############################################################################
#
# PUBLIC API
#
# ############################################################################

class InfoAPIView(APIView):
    """Public API for GeoKey server information."""

    def get(self, request):
        """
        Handle GET request.

        Return GeoKey server information.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.

        Returns
        -------
        rest_framework.response.Response
            Contains the GeoKey server information.
        """
        info = {'geokey': {}}
        info['geokey']['version'] = get_version()
        info['geokey']['installed_extensions'] = map(
            lambda (ext_id, ext): {
                'name': ext_id,
                'version': ext['version'] if 'version' in ext else None
            },
            extensions.iteritems()
        )

        return Response(info)
